# vim:ts=4 sw=4 expandtab softtabstop=4
from jsonmerge.exceptions import HeadInstanceError, SchemaError
from jsonmerge.jsonvalue import JSONValue

class Descender(object):
    """Base class for descender classes.

    Descenders are similar to merge strategies, except that they only handle
    recursion into deeper schema structures and don't touch instances.
    """
    def descend_instance(self, walk, schema, base, head):
        return None

    def descend_schema(self, walk, schema):
        return None

class Ref(Descender):
    def __init__(self):
        self.refs_descended = set('#')

    def descend_instance(self, walk, schema, base, head):
        ref = schema.val.get("$ref")
        if ref is None:
            return None

        with walk.resolver.resolving(ref) as resolved:
            return walk.descend(JSONValue(resolved, ref), base, head)

    def descend_schema(self, walk, schema):
        ref = schema.val.get("$ref")
        if ref is None:
            return None

        if ref in self.refs_descended:
            return schema

        if walk.resolver.is_remote_ref(ref):
            return schema

        self.refs_descended.add(ref)

        with walk.resolver.resolving(ref) as resolved:

            rinstance = JSONValue(resolved, ref)
            if not walk.is_type(rinstance, 'object'):
                raise SchemaError("'$ref' does not point to an object", schema)

            result = walk.descend(rinstance)

            resolved.clear()
            resolved.update(result.val)

        return schema

class OneOf(Descender):
    def do_descend(self, schema):
        one_of = schema.get("oneOf")
        if one_of.is_undef():
            return False

        # If we have a strategy defined on this level, don't descend into
        # subschemas.
        if not schema.get("mergeStrategy").is_undef():
            return False

        return True

    def descend_instance(self, walk, schema, base, head):
        if not self.do_descend(schema):
            return None

        one_of = schema.get("oneOf")

        valid = []

        def is_valid(v, schema):
            if v.is_undef():
                return True
            else:
                return not list(walk.merger.validator.iter_errors(v.val, schema))

        for i, subschema in enumerate(one_of.val):
            base_valid = is_valid(base, subschema)
            head_valid = is_valid(head, subschema)

            if base_valid and head_valid:
                valid.append(i)

        if len(valid) == 0:
            raise HeadInstanceError("No element of 'oneOf' validates both base and head", head)

        if len(valid) > 1:
            raise HeadInstanceError("Multiple elements of 'oneOf' validate", head)

        i = valid[0]
        return walk.descend(one_of[i], base, head)

    def descend_schema(self, walk, schema):
        if not self.do_descend(schema):
            return None

        one_of = schema.get("oneOf")

        for i in range(len(one_of.val)):
            one_of[i] = walk.descend(one_of[i])

        return schema

class AnyOfAllOf(Descender):
    def descend(self, schema):
        allOf = schema.get("allOf")
        anyOf = schema.get("anyOf")
        if allOf.is_undef() and anyOf.is_undef():
            return None

        # We must have a strategy defined on this level, or we can't know which
        # subschema to descend to.
        if not schema.get("mergeStrategy").is_undef():
            return None

        raise SchemaError("Can't descend to 'allOf' and 'anyOf' keywords", schema)

    def descend_instance(self, walk, schema, base, head):
        return self.descend(schema)

    def descend_schema(self, walk, schema):
        return self.descend(schema)



class JSONMergeError(Exception):
	def __init__(self, message, value=None, strategy_name=None):
		self.message = message
		self.value = value
		self.strategy_name = strategy_name

	def __str__(self):
		r = self.message

		try:
			ref = self.value.ref
		except:
			ref = None

		if ref is not None:
			r = "%s: %s" % (r, ref)

		if self.strategy_name is not None:
			r = "'%s' merge strategy: %s" % (self.strategy_name, r)

		return r

class BaseInstanceError(JSONMergeError): pass
class HeadInstanceError(JSONMergeError): pass
class SchemaError(JSONMergeError): pass



# vim:ts=4 sw=4 expandtab softtabstop=4
from jsonschema.validators import RefResolver, urldefrag, urljoin

class LocalRefResolver(RefResolver):
    # We want to have a class that is the same as jsonschema's RefResolver
    # except:
    #
    #  * No caching for resolves. We are changing the schema as we walk through
    #    it, so with caching you can get outdated resolves.
    #
    #  * Provide a _is_remote_ref() method to check if a $ref points to an
    #    external reference.

    def __init__(self, *args, **kwargs):
        try:
            kwargs["remote_cache"] = self.resolve_from_url
        except AttributeError:
            # jsonschema<=2.4.0
            pass
        super(LocalRefResolver, self).__init__(*args, **kwargs)

    def is_remote_ref(self, ref):
        url = urljoin(self.resolution_scope, ref)
        url, fragment = urldefrag(url)
        return url != self.base_uri



# vim:ts=4 sw=4 expandtab softtabstop=4
import sys

if sys.version_info[0] >= 3:
    text_type = str
else:
    text_type = unicode

class JSONValue(object):
    def __init__(self, val=None, ref='#', undef=False):
        assert not isinstance(val, JSONValue)
        self.val = val
        self.ref = ref
        self.undef = undef

    def is_undef(self):
        return self.undef

    def _ref_escape(self, key):
        return key.replace('~', '~0').replace('/', '~1')

    def _subval(self, key, **kwargs):
        return JSONValue(ref=self.ref+'/'+self._ref_escape(text_type(key)), **kwargs)

    def __setitem__(self, key, item):
        if item.is_undef():
            if isinstance(self.val, list):
                raise ValueError("Can't assign an undefined value to a list")

            # setting a dict element to an undefined value deletes that element
            if key in self.val:
                del self.val[key]
        else:
            self.val[key] = item.val

    def __getitem__(self, key):
        return self._subval(key, val=self.val[key])

    def append(self, item):
        assert isinstance(self.val, list)

        if not item.is_undef():
            self.val.append(item.val)

    def get(self, key, *args):
        r = self.val.get(key, *args)
        if r is None:
            return self._subval(key, undef=True)
        else:
            return self._subval(key, val=r)

    def __repr__(self):
        if self.is_undef():
            return 'JSONValue(undef=True)'
        else:
            return 'JSONValue(%r,%r)' % (self.val, self.ref)

    def items(self):
        for k, v in self.val.items():
            yield (k, self._subval(k, val=v))

    def __iter__(self):
        assert isinstance(self.val, list)

        for i, v in enumerate(self.val):
            yield self._subval(i, val=v)



# vim:ts=4 sw=4 expandtab softtabstop=4
import unittest
from jsonmerge.jsonvalue import JSONValue

class TestJSONValue(unittest.TestCase):

    def test_get_attr(self):
        v = JSONValue({'a': 'b'})

        va = v['a']
        self.assertEqual('b', va.val)
        self.assertEqual('#/a', va.ref)

    def test_get_attr_nonascii(self):
        v = JSONValue({u'\u20ac': 'b'})

        va = v[u'\u20ac']
        self.assertEqual('b', va.val)
        self.assertEqual(u'#/\u20ac', va.ref)

    def test_get_attr_escape_slash(self):
        v = JSONValue({'a/b': 'c'})

        va = v['a/b']
        self.assertEqual('c', va.val)
        self.assertEqual('#/a~1b', va.ref)

    def test_get_attr_escape_tilde(self):
        v = JSONValue({'~0': 'a'})

        va = v['~0']
        self.assertEqual('a', va.val)
        self.assertEqual('#/~00', va.ref)

    def test_get_default(self):
        v = JSONValue({})

        va = v.get('a')
        self.assertTrue(va.is_undef())

    def test_get(self):
        v = JSONValue({'a': 'b'})

        va = v.get('a')
        self.assertTrue('b', va.val)
        self.assertEqual('#/a', va.ref)

    def test_undef(self):
        v = JSONValue(undef=True)
        self.assertTrue(v.is_undef())

    def test_dict_set_attr(self):
        v = JSONValue({})

        v['a'] = JSONValue('b')

        self.assertEqual({'a': 'b'}, v.val)

    def test_dict_set_attr_undef(self):
        v = JSONValue({})

        v['a'] = JSONValue(undef=True)

        self.assertEqual({}, v.val)

    def test_dict_set_attr_undef_exists(self):
        v = JSONValue({'a': 'b'})

        v['a'] = JSONValue(undef=True)

        self.assertEqual({}, v.val)

    def test_list_set_attr(self):
        v = JSONValue([1])

        v[0] = JSONValue(2)

        self.assertEqual([2], v.val)

    def test_list_set_attr_undef_exists(self):
        v = JSONValue([1])

        # Setting a list element to undef does not make sense.
        with self.assertRaises(ValueError) as cm:
            v[0] = JSONValue(undef=True)

    def test_list_set_attr_undef(self):
        v = JSONValue([])

        with self.assertRaises(ValueError) as cm:
            v[0] = JSONValue(undef=True)

    def test_append_undef(self):
        v = JSONValue([])

        v.append(JSONValue(undef=True))

        self.assertEqual([], v.val)
